(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: Water.cpp
//
// Desc:
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit WaterUnit;

interface

uses
  Windows, Messages, SysUtils, Resource, 
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

{.$DEFINE DEBUG_VS}   // Uncomment this line to debug vertex shaders
{.$DEFINE DEBUG_PS}   // Uncomment this line to debug pixel shaders

// C++Builder compile hack
{$NOINCLUDE Resource}
{$HPPEMIT '#include "Resource.hpp"'}

const
  WATER_DEPTH = 20.0;
  WATER_CAUSTICS_SIZE = 128;

type
  CEnvironment = class
  public
    constructor Create;
    destructor Destroy; override;

    function Initialize(fSize: Single): HResult;

    function OnCreateDevice(const pDevice: IDirect3DDevice9): HResult;
    function OnResetDevice: HResult;
    function OnLostDevice: HResult;
    function OnDestroyDevice: HResult;

    function SetSurfaces( const
        pXNeg, pXPos,
        pYNeg, pYPos,
        pZNeg, pZPos: IDirect3DTexture9): HResult;

    function Draw: HResult;

  protected
    m_fSize: Single;

    m_pDevice: IDirect3DDevice9;
    m_pSurf: array[0..5] of IDirect3DTexture9;
  end;

  TWaterRefract = packed record
    // Vrefract = (V + refract * N) * norm
    fRefract:     Single;
    fRefractNorm: Single;
    dwDiffuse:    DWord;
  end;
  PWaterRefractArray = ^TWaterRefractArray;
  TWaterRefractArray = array [0..0] of TWaterRefract;

  TWaterSurface = packed record
    fHeight:      Single;
    fVelocity:    Single;
  end;
  PWaterSurfaceArray = ^TWaterSurfaceArray;
  TWaterSurfaceArray = array [0..0] of TWaterSurface;

//-----------------------------------------------------------------------------
// Name: CWater
// Desc: The water that is rendered
//-----------------------------------------------------------------------------
  CWater = class
  private
    m_fSize:            Single;
    m_fDepth:           Single;
    m_fScaleTex:        Single;
    m_fAvgHeight:       Single;

// These vars are not used on original example code :BAA
//    m_fSphereHeight:   Single;
//    m_fSphereRadius2:  Single;

    m_uIndices:         LongWord;
    m_uVertices:        LongWord;

    m_pSurface:         PWaterSurfaceArray;
    m_pRefract:         PWaterRefractArray;

    m_pDevice:          IDirect3DDevice9;
    m_pibIndices:       IDirect3DIndexBuffer9;
    m_pvbVertices:      IDirect3DVertexBuffer9;
    m_pvbCaustics:      IDirect3DVertexBuffer9;
  public
    constructor Create;
    destructor Destroy; override;

    function Initialize(fSize, fDepth: Single): HResult;

    function OnCreateDevice(const pDevice: IDirect3DDevice9): HResult;
    function OnResetDevice: HResult;
    function OnLostDevice: HResult;
    function OnDestroyDevice: HResult;

    function Drop: HResult;
    function Update(vecPos, vecLight: TD3DXVector3; bCalcCaustics: BOOL): HResult;

    function DrawCaustics: HResult;
    function DrawSurface: HResult;
  end;


//-----------------------------------------------------------------------------
// Name: CMyD3DApplication
// Desc:
//-----------------------------------------------------------------------------
type
  CMyD3DApplication = class(CD3DApplication)
  public
    constructor Create;
    destructor Destroy; override;
    function GetNextTechnique(nDir: Integer; bBypassValidate: BOOL): HRESULT;

    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function FrameMove: HResult; override;
    function Render: HResult; override;
    function FinalCleanup: HResult; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;

    function MsgProc(hwnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  private
    m_Water:                CWater;
    m_Environment:          CEnvironment;

    m_pEffect:              ID3DXEffect;
    m_iTechnique:           LongWord;
    m_pRenderToSurface:     ID3DXRenderToSurface;

    m_vecLight:             TD3DXVector4;
    m_colorLight:           TD3DXColor;

    m_pFloorTex:            IDirect3DTexture9;
    m_pCausticTex:          IDirect3DTexture9;
    m_pCausticSurf:         IDirect3DSurface9;
    m_pSkyTex:              array[0..5] of IDirect3DTexture9;
    m_pSkyCubeTex:          IDirect3DCubeTexture9;

    m_pFont:                CD3DFont;
    m_pFontSmall:           CD3DFont;

    m_bKey:                 array[0..255] of Boolean;

    m_bPause:               BOOL;
    m_bDrawWater:           BOOL;
    m_bDrawCaustics:        BOOL;
    m_bDrawEnvironment:     BOOL;
    m_bShowHelp:            BOOL;

    m_fSpeed:               Single;
    m_fAngularSpeed:        Single;
    m_fTime:                Single;
    m_fSecsPerFrame:        Single;

    //todo: fill bug report - "should be  'BOOL m_bAutoAddDrop'"
    m_bAutoAddDrop:         Boolean;
    m_fAddDropTimer:        Single;
    m_fAddDropCountdown:    Single;

    m_vecVelocity:          TD3DXVector3;
    m_vecAngularVelocity:   TD3DXVector3;

    m_matIdentity:          TD3DXMatrix;
    m_matView:              TD3DXMatrix;
    m_matPosition:          TD3DXMatrix;
    m_matProjection:        TD3DXMatrix;
  end;


implementation

uses Math;

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication
// Desc:
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create();
var
  i: Integer;
begin
  inherited Create;

  m_strWindowTitle    := 'Water';
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_pFont         := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall    := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);

  // Water
  m_pEffect       := nil;
  m_iTechnique    := 0;

  m_vecLight      := D3DXVECTOR4(0.5, -1.0, 1.0, 0.0);
  m_colorLight    := D3DXCOLOR(1.0, 1.0, 1.0, 1.0);

  m_pFloorTex     := nil;
  m_pCausticTex   := nil;
  m_pCausticSurf  := nil;
  m_pSkyCubeTex   := nil;

  for i:= 0 to 5 do m_pSkyTex[i]:= nil;

  // Misc
  FillChar(m_bKey, SizeOf(m_bKey), False);

  m_fSpeed        := 25.0;
  m_fAngularSpeed := 1.0;
  m_fSecsPerFrame := 0.0;
  m_fTime         := 0.0;

  m_bAutoAddDrop  := True;
  m_fAddDropCountdown := 0.5;
  m_fAddDropTimer := 0.0;

  m_vecVelocity        := D3DXVector3(0.0, 0.0, 0.0);
  m_vecAngularVelocity := D3DXVector3(0.0, 0.0, 0.0);

  D3DXMatrixIdentity(m_matIdentity);
  D3DXMatrixIdentity(m_matView);
  D3DXMatrixIdentity(m_matPosition);
  D3DXMatrixIdentity(m_matProjection);

  m_bShowHelp         := FALSE;
  m_bPause            := FALSE;

  m_Water:= CWater.Create;
  m_Environment:= CEnvironment.Create;
end;

destructor CMyD3DApplication.Destroy;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pFontSmall);
  inherited;
end;

//-----------------------------------------------------------------------------
// Name: GetNextTechnique
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.GetNextTechnique(nDir: Integer; bBypassValidate: BOOL): HRESULT;
var
  effect:      TD3DXEffectDesc;
  iTechnique: Longword;
  szText: array[0..255] of Char;
begin
  iTechnique:= m_iTechnique;

  m_pEffect.GetDesc(effect);

  while True do
  begin
    iTechnique:= iTechnique + DWord(nDir);

    // cast is needed because iTechnique is unsigned, so always >= 0
    if (Integer(iTechnique) < 0) then iTechnique:= effect.Techniques - 1;

    if (iTechnique >= effect.Techniques) then iTechnique:= 0;

    if (nDir <> 0) and (iTechnique = m_iTechnique) then Break;

    if (nDir = 0) then nDir:= 1;

    m_pEffect.SetTechnique(m_pEffect.GetTechnique(iTechnique));

    if (bBypassValidate or (iTechnique = effect.Techniques - 1) or
       (m_bDrawCaustics or
          not m_pEffect.IsParameterUsed('tCAU', nil)) and SUCCEEDED(m_pEffect.ValidateTechnique(nil))) then
    begin
      m_iTechnique := iTechnique;

      StrFmt(szText, 'Water - Technique %d', [m_iTechnique]);
      SetWindowText(m_hWnd, szText);

      Result:= S_OK;
      Exit;
    end;
  end;

  m_pEffect.SetTechnique(m_pEffect.GetTechnique(m_iTechnique));
  Result:= E_FAIL;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  // Initialize Water
  Result:= m_Water.Initialize(64.0, WATER_DEPTH);
  if FAILED(Result) then Exit;

  // Initialize Environment
  Result:= m_Environment.Initialize(1000.0);
  if FAILED(Result) then Exit;

  // Misc stuff
  m_matPosition._11 := 1.0;
  m_matPosition._22 := 0.941739;
  m_matPosition._23 := -0.33634159;
  m_matPosition._32 := 0.33634159;
  m_matPosition._41 := 1.3352206;
  m_matPosition._42 := 6.2386379;
  m_matPosition._43 := 18.534956;
  m_matPosition._44 := 1.0;

  D3DXMatrixInverse(m_matView, nil, m_matPosition);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  sz: array[0..511] of Char;
  i: Integer;
  dwShaderFlags: DWORD;

const
  // Sky
  szSkyTex: array[0..5] of PChar = (
    'lobbyxpos.jpg', 'lobbyxneg.jpg',
    'lobbyypos.jpg', 'lobbyyneg.jpg',
    'lobbyzpos.jpg', 'lobbyzneg.jpg'
  );

begin
  m_bDrawWater       := True;
  m_bDrawCaustics    := True;
  m_bDrawEnvironment := True;


  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  // Floor
  Result := DXUtil_FindMediaFileCb(sz, SizeOf(sz), 'Water.bmp');
  if FAILED(Result) then Exit;
  D3DXCreateTextureFromFileEx(m_pd3dDevice, sz, D3DX_DEFAULT, D3DX_DEFAULT,
      D3DX_DEFAULT, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED, D3DX_DEFAULT,
      D3DX_DEFAULT, 0, nil, nil, m_pFloorTex);


  for i:= 0 to 5 do
  begin
    Result := DXUtil_FindMediaFileCb(sz, SizeOf(sz), szSkyTex[i]);
    if FAILED(Result) then Exit;
    D3DXCreateTextureFromFileEx(m_pd3dDevice, sz, D3DX_DEFAULT, D3DX_DEFAULT,
        1, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED, D3DX_DEFAULT, D3DX_DEFAULT,
        0, nil, nil, m_pSkyTex[i]);
  end;

  if (m_d3dCaps.TextureCaps and D3DPTEXTURECAPS_CUBEMAP <> 0) then
  begin
    Result := DXUtil_FindMediaFileCb(sz, SizeOf(sz), 'LobbyCube.dds');
    if FAILED(Result) then Exit;
    Result := D3DXCreateCubeTextureFromFile(m_pd3dDevice, sz, m_pSkyCubeTex);
    if FAILED(Result) then Exit;
  end;

  // OnCreateDevice
  Result := m_Water.OnCreateDevice(m_pd3dDevice);
  if FAILED(Result) then Exit;

  Result := m_Environment.OnCreateDevice(m_pd3dDevice);
  if FAILED(Result) then Exit;

  // Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the shader debugger.
  // Debugging vertex shaders requires either REF or software vertex processing, and debugging
  // pixel shaders requires REF.  The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug
  // experience in the shader debugger.  It enables source level debugging, prevents instruction
  // reordering, prevents dead code elimination, and forces the compiler to compile against the next
  // higher available software target, which ensures that the unoptimized shaders do not exceed
  // the shader model limitations.  Setting these flags will cause slower rendering since the shaders
  // will be unoptimized and forced into software.  See the DirectX documentation for more information
  // about using the shader debugger.
  dwShaderFlags := 0;
  {$IFDEF DEBUG_VS}
    dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
  {$ENDIF}

  // Effect
  DXUtil_FindMediaFileCb(sz, SizeOf(sz), 'water.fx');

  Result := D3DXCreateEffectFromFile(m_pd3dDevice, sz, nil, nil, dwShaderFlags, nil, m_pEffect, nil);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  light: TD3DLight9;
  material: TD3Dmaterial9;
  mode: TD3DDisplayMode;
  desc: TD3DSurfaceDesc;
begin
  // Restore the font
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  // Create light
  FillChar(light, SizeOf(light), 0);

  light._Type      := D3DLIGHT_DIRECTIONAL;
  light.Diffuse.r  := m_colorLight.r;
  light.Diffuse.g  := m_colorLight.g;
  light.Diffuse.b  := m_colorLight.b;
  light.Diffuse.a  := m_colorLight.a;
  light.Specular.r := 1.0;
  light.Specular.g := 1.0;
  light.Specular.b := 1.0;
  light.Specular.a := 0.0;
  light.Direction.x:= m_vecLight.x;
  light.Direction.y:= m_vecLight.y;
  light.Direction.z:= m_vecLight.z;

  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);


  // Create material
  FillChar(material, SizeOf(material), 0);

  material.Diffuse.a := 1.0;
  material.Specular.r:= 0.5;
  material.Specular.g:= 0.5;
  material.Specular.b:= 0.5;
  material.Power     := 20.0;

  m_pd3dDevice.SetMaterial(material);

  // Setup render states
  m_pd3dDevice.SetFVF(D3DFVF_XYZ);

  m_pd3dDevice.SetTransform(D3DTS_VIEW,  m_matView);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matIdentity);

  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,       iFalse); // FALSE
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse); // FALSE

  m_pd3dDevice.SetRenderState(D3DRS_ZFUNC,     D3DCMP_LESSEQUAL);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,  D3DCULL_CW);
  m_pd3dDevice.SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_ONE);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_CURRENT);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);

  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAARG2, D3DTA_CURRENT);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_POINT);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 0);

  // Create caustic texture
  m_pd3dDevice.GetDisplayMode(0, mode);

  if (FAILED(D3DXCreateTexture(m_pd3dDevice, WATER_CAUSTICS_SIZE, WATER_CAUSTICS_SIZE, 1, D3DUSAGE_RENDERTARGET, mode.Format, D3DPOOL_DEFAULT, m_pCausticTex))) then
  begin
    Result:= D3DXCreateTexture(m_pd3dDevice, WATER_CAUSTICS_SIZE, WATER_CAUSTICS_SIZE, 1, 0, mode.Format, D3DPOOL_DEFAULT, m_pCausticTex);
    if FAILED(Result) then Exit;
  end;

  m_pCausticTex.GetSurfaceLevel(0, m_pCausticSurf);
  m_pCausticSurf.GetDesc(desc);

  Result:= D3DXCreateRenderToSurface(m_pd3dDevice, desc.Width, desc.Height,
    desc.Format, FALSE, D3DFMT_UNKNOWN, m_pRenderToSurface);
  if FAILED(Result) then Exit;

  // Effect
  m_pEffect.OnResetDevice;

  m_pEffect.SetMatrix('mID',  m_matIdentity);
  m_pEffect.SetMatrix('mENV', m_matIdentity);

  m_pEffect.SetTexture('tFLR', m_pFloorTex);
  m_pEffect.SetTexture('tCAU', m_pCausticTex);
  m_pEffect.SetTexture('tENV', m_pSkyCubeTex);

  Result:= GetNextTechnique(0, FALSE);
  if FAILED(Result) then Exit;

  // Set surfaces
  Result:= m_Environment.SetSurfaces(
      m_pSkyTex[Integer(D3DCUBEMAP_FACE_NEGATIVE_X)], m_pSkyTex[Integer(D3DCUBEMAP_FACE_POSITIVE_X)],
      m_pSkyTex[Integer(D3DCUBEMAP_FACE_NEGATIVE_Y)], m_pSkyTex[Integer(D3DCUBEMAP_FACE_POSITIVE_Y)],
      m_pSkyTex[Integer(D3DCUBEMAP_FACE_POSITIVE_Z)], m_pSkyTex[Integer(D3DCUBEMAP_FACE_NEGATIVE_Z)]);
  if FAILED(Result) then Exit;

  Result:= m_Water.OnResetDevice;
  if FAILED(Result) then Exit;

  Result:= m_Environment.OnResetDevice;
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;

  m_Water.OnLostDevice;
  m_Environment.OnLostDevice;
  m_pEffect.OnLostDevice;

  SAFE_RELEASE(m_pRenderToSurface);
  SAFE_RELEASE(m_pCausticSurf);
  SAFE_RELEASE(m_pCausticTex);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
var
  i: Integer;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;

  m_Water.OnDestroyDevice;
  m_Environment.OnDestroyDevice;

  SAFE_RELEASE(m_pFloorTex);
  SAFE_RELEASE(m_pSkyCubeTex);
  SAFE_RELEASE(m_pEffect);

  for i:= 0 to 5 do
    m_pSkyTex[i]:= nil;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  vecT: TD3DXVector3;
  vecR: TD3DXVector3;
  v1, v2: TD3DXVector3; // temporary
  matT, matR: TD3DXMatrix;
  qR: TD3DXQuaternion;

  bCaustics: BOOL;
  vecPos: TD3DXVector3;
  vecLight: TD3DXVector3;

  matProj: TD3DXMatrix;
  matView: TD3DXMatrix;
begin
  //
  // Process keyboard input
  //

  vecT:= D3DXVector3Zero;
  vecR:= D3DXVector3Zero;

  if (m_bKey[VK_NUMPAD1] or m_bKey[VK_LEFT])   then vecT.x:= vecT.x - 1.0; // Slide Left
  if (m_bKey[VK_NUMPAD3] or m_bKey[VK_RIGHT])  then vecT.x:= vecT.x + 1.0; // Slide Right
  if (m_bKey[VK_DOWN])                         then vecT.y:= vecT.y - 1.0; // Slide Down
  if (m_bKey[VK_UP])                           then vecT.y:= vecT.y + 1.0; // Slide Up
  if (m_bKey[Ord('W')])                        then vecT.z:= vecT.z - 2.0; // Move Forward
  if (m_bKey[Ord('S')])                        then vecT.z:= vecT.z + 2.0; // Move Backward
  if (m_bKey[Ord('A')] or m_bKey[VK_NUMPAD8])  then vecR.x:= vecR.x - 1.0; // Pitch Down
  if (m_bKey[Ord('Z')] or m_bKey[VK_NUMPAD2])  then vecR.x:= vecR.x + 1.0; // Pitch Up
  if (m_bKey[Ord('E')] or m_bKey[VK_NUMPAD6])  then vecR.y:= vecR.y - 1.0; // Turn Right
  if (m_bKey[Ord('Q')] or m_bKey[VK_NUMPAD4])  then vecR.y:= vecR.y + 1.0; // Turn Left
  if (m_bKey[VK_NUMPAD9])                      then vecR.z:= vecR.z - 2.0; // Roll CW
  if (m_bKey[VK_NUMPAD7])                      then vecR.z:= vecR.z + 2.0; // Roll CCW

  // m_vecVelocity:= m_vecVelocity * 0.9 + vecT * 0.1;
  D3DXVec3Add(m_vecVelocity,
    D3DXVec3Scale(v1, m_vecVelocity, 0.9)^,
    D3DXVec3Scale(v2, vecT, 0.1)^);
  // m_vecAngularVelocity:= m_vecAngularVelocity * 0.9 + vecR * 0.1s;
  D3DXVec3Add(m_vecAngularVelocity,
    D3DXVec3Scale(v1, m_vecAngularVelocity, 0.9)^,
    D3DXVec3Scale(v2, vecR, 0.1)^);

  if m_bAutoAddDrop then
  begin
    m_fAddDropTimer := m_fAddDropTimer - m_fElapsedTime;
    if (m_fAddDropTimer < 0.0) then
    begin
      m_fAddDropTimer := m_fAddDropCountdown;
      m_Water.Drop;
    end;
  end;

  //
  // Update position and view matricies
  //

  // vecT:= m_vecVelocity * m_fElapsedTime * m_fSpeed;
  D3DXVec3Scale(vecT, m_vecVelocity, m_fElapsedTime * m_fSpeed);
  // vecR:= m_vecAngularVelocity * m_fElapsedTime * m_fAngularSpeed;
  D3DXVec3Scale(vecR, m_vecAngularVelocity, m_fElapsedTime * m_fAngularSpeed);

  D3DXMatrixTranslation(matT, vecT.x, vecT.y, vecT.z);
  D3DXMatrixMultiply(m_matPosition, matT, m_matPosition);

  D3DXQuaternionRotationYawPitchRoll(qR, vecR.y, vecR.x, vecR.z);
  D3DXMatrixRotationQuaternion(matR, qR);

  D3DXMatrixMultiply(m_matPosition, matR, m_matPosition);
  D3DXMatrixInverse(m_matView, nil, m_matPosition);


  //
  // Update simulation
  //

  if (not m_bPause and m_bDrawWater) then
  begin
    bCaustics:= m_bDrawCaustics and m_pEffect.IsParameterUsed('tCAU', nil);
    vecPos:= D3DXVector3(m_matPosition._41, m_matPosition._42, m_matPosition._43);
    vecLight:= D3DXVector3(0.0, 1.0, 0.0);

    m_Water.Update(vecPos, vecLight, bCaustics);
    m_fTime:= m_fTime + m_fSecsPerFrame;

    if (bCaustics) then
    begin
      if SUCCEEDED(m_pRenderToSurface.BeginScene(m_pCausticSurf, nil)) then
      begin
        D3DXMatrixOrthoRH(matProj, 63.0, 63.0, 1.0, 100.0);
        D3DXMatrixRotationX(matView, 0.5 * D3DX_PI);
        matView._43:= -50.0;

        m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);
        m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

        m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, $ff000000, 0.0, 0);

        m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);
        m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
        m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue); // TRUE

        m_Water.DrawCaustics;

        m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse); // FALSE
        m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
        m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);

        m_pRenderToSurface.EndScene(0);
      end else
      begin
        m_bDrawCaustics:= FALSE;
        m_pEffect.SetTexture('tCAU', nil);

        Result:= GetNextTechnique(0, False);
        if FAILED(Result) then Exit;
      end;
    end;
  end;

  Result:= S_OK;
end;

function CMyD3DApplication.Render: HResult;
var
  fAspectRatio: Single;

  matP: TD3DXMatrix;
  mat: TD3DXMatrix;
  uPasses: Longword;
  uPass: Longword;

  sText: String;
begin
  Result:= m_pd3dDevice.BeginScene;
  if FAILED(Result) then Exit;

  // Draw Environment
  fAspectRatio:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovRH(m_matProjection, D3DXToRadian(60.0), fAspectRatio, 0.1, 2000.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProjection);

  if m_bDrawEnvironment then
  begin
    mat:= m_matView;
    mat._41:= 0.0; mat._42:= 0.0; mat._43:= 0.0;
    m_pd3dDevice.SetTransform(D3DTS_VIEW, mat);

    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
    m_pd3dDevice.SetRenderState(D3DRS_ZFUNC, D3DCMP_ALWAYS);

    m_Environment.Draw;

    m_pd3dDevice.SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);
  end else
  begin
    m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $00000000, 1.0, 0);
  end;

  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);

  // Draw water
  if m_bDrawWater then
  begin
    // Setup matrices
    if (m_pEffect.IsParameterUsed('mENV', nil)) then
    begin
      matP:= m_matPosition;
      matP._41:= 0.0; matP._42:= 0.0; matP._43:= 0.0;

      D3DXMatrixScaling(mat, 1.0, 1.0, -1.0);

      D3DXMatrixMultiply(mat, matP, mat);

      // matCube
      m_pEffect.SetMatrix('mENV', mat);
    end;

    // Draw water
    m_pEffect._Begin(@uPasses, 0);

    for uPass:= 0 to uPasses - 1 do
    begin
      m_pEffect.Pass(uPass);
      m_Water.DrawSurface;
    end;

    m_pEffect._End;
  end;

  // Show info
  m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

  sText:= Format('Using Technique %d', [m_iTechnique]);
  m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,255,100,100), PChar(sText));

  if m_bShowHelp then
  begin
    m_pFontSmall.DrawText(2, 60, D3DCOLOR_ARGB(255,100,100,200),
                          'Keyboard controls:' );
    m_pFontSmall.DrawText(20, 80, D3DCOLOR_ARGB(255,100,100,200),
                          'Add Drop'#10 +
                          'Next Technique'#10 +
                          'Next Tech. (no validate)'#10 +
                          'Prev Technique'#10 +
                          'Prev Tech. (no validate)'#10 +
                          'Move'#10'Turn'#10'Pitch'#10'Slide'#10 +
                          'Help'#10'Change device'#10'Exit');
    m_pFontSmall.DrawText(210, 80, D3DCOLOR_ARGB(255,100,100,200),
                          'D'#10 +
                          'PageDn'#10'Shift-PageDn'#10 +
                          'PageUp'#10'Shift-PageUp'#10 +
                          'W,S'#10'E,Q'#10'A,Z'#10'Arrow keys'#10 +
                          'F1'#10'F2'#10'Esc');
  end else
  begin
    m_pFontSmall.DrawText( 2, 60, D3DCOLOR_ARGB(255,100,100,200),
                          'Press F1 for help' );
  end;

  Result:= m_pd3dDevice.EndScene;
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFontSmall);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // Debugging vertex shaders requires either REF or software vertex processing
  // and debugging pixel shaders requires REF.
  {$IFDEF DEBUG_VS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF) and
       ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0)
    then Exit;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF)
    then Exit;
  {$ENDIF}

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hwnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT;
begin
  case uMsg of
    WM_KEYDOWN: m_bKey[wParam]:= True;
    WM_KEYUP:   m_bKey[wParam]:= False;
    WM_COMMAND:
     case LOWORD(wParam) of
       IDM_ADDDROP:
       begin
         m_bAutoAddDrop:= False;
         m_Water.Drop;
       end;
       IDM_NEXT_TECHNIQUE:              GetNextTechnique(1, False);
       IDM_NEXT_TECHNIQUE_NOVALIDATE:   GetNextTechnique(1, True);
       IDM_PREV_TECHNIQUE:              GetNextTechnique(-1, False);
       IDM_PREV_TECHNIQUE_NOVALIDATE:   GetNextTechnique(-1, True);
       IDM_TOGGLEHELP:                  m_bShowHelp:= not m_bShowHelp;
     end;
  end; // case

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;


//////////////////////////////////////////////////////////////////////////////
// Types /////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

type
  TEnvVertex = packed record
    m_vecPos: TD3DXVector3;
    m_vecTex: TD3DXVector2;
  end;

const
  // #define ENV_FVF (D3DFVF_XYZ | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0))
  TEnvVertex_FVF = D3DFVF_XYZ or D3DFVF_TEX1 or D3DFVF_TEXTUREFORMAT2;


//////////////////////////////////////////////////////////////////////////////
// CEnvironment implementation ///////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


constructor CEnvironment.Create;
begin
  m_fSize:= 1.0;
  m_pDevice:= nil;

  FillChar(m_pSurf, SizeOf(m_pSurf), $00);
end;

destructor CEnvironment.Destroy;
begin
  inherited;
end;

function CEnvironment.Initialize(fSize: Single): HResult;
begin
  m_fSize:= fSize;
  Result:= S_OK;
end;

function CEnvironment.OnCreateDevice(const pDevice: IDirect3DDevice9): HResult;
begin
  m_pDevice:= pDevice;
  Result:= S_OK;
end;

function CEnvironment.OnResetDevice: HResult;
begin
  Result:= S_OK;
end;

function CEnvironment.OnLostDevice: HResult;
var
  i: Integer;
begin
  for i:= 0 to 5 do m_pSurf[i] := nil;
  Result:= S_OK;
end;

function CEnvironment.OnDestroyDevice: HResult;
begin
  m_pDevice:= nil;
  Result:= S_OK;
end;

function CEnvironment.SetSurfaces(const
  pXNeg, pXPos,
  pYNeg, pYPos,
  pZNeg, pZPos: IDirect3DTexture9): HResult;
begin
  m_pSurf[0]:= pXNeg;
  m_pSurf[1]:= pXPos;
  m_pSurf[2]:= pYNeg;
  m_pSurf[3]:= pYPos;
  m_pSurf[4]:= pZNeg;
  m_pSurf[5]:= pZPos;

  Result:= S_OK;
end;

function CEnvironment.Draw: HResult;
var
  f: Single;
  vert: array[0..3] of TEnvVertex;
begin
  f:= 0.5 / 512.0;

  vert[0].m_vecTex:= D3DXVector2(0.0 + f, 0.0 + f);
  vert[1].m_vecTex:= D3DXVector2(0.0 + f, 1.0 - f);
  vert[2].m_vecTex:= D3DXVector2(1.0 - f, 0.0 + f);
  vert[3].m_vecTex:= D3DXVector2(1.0 - f, 1.0 - f);

  m_pDevice.SetFVF(TEnvVertex_FVF);
  f:= m_fSize * 0.5;

  // XNeg
  vert[0].m_vecPos:= D3DXVector3(-f,  f,  f);
  vert[1].m_vecPos:= D3DXVector3(-f, -f,  f);
  vert[2].m_vecPos:= D3DXVector3(-f,  f, -f);
  vert[3].m_vecPos:= D3DXVector3(-f, -f, -f);

  m_pDevice.SetTexture(0, m_pSurf[0]);
  m_pDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, vert, SizeOf(TEnvVertex));

  // XPos
  vert[0].m_vecPos:= D3DXVector3(f,  f, -f);
  vert[1].m_vecPos:= D3DXVector3(f, -f, -f);
  vert[2].m_vecPos:= D3DXVector3(f,  f,  f);
  vert[3].m_vecPos:= D3DXVector3(f, -f,  f);

  m_pDevice.SetTexture(0, m_pSurf[1]);
  m_pDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, vert, SizeOf(TEnvVertex));

  // YNeg
  vert[0].m_vecPos:= D3DXVector3(-f, -f, -f);
  vert[1].m_vecPos:= D3DXVector3(-f, -f,  f);
  vert[2].m_vecPos:= D3DXVector3( f, -f, -f);
  vert[3].m_vecPos:= D3DXVector3( f, -f,  f);

  m_pDevice.SetTexture(0, m_pSurf[2]);
  m_pDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, vert, SizeOf(TEnvVertex));

  // YPos
  vert[0].m_vecPos:= D3DXVector3(-f,  f,  f);
  vert[1].m_vecPos:= D3DXVector3(-f,  f, -f);
  vert[2].m_vecPos:= D3DXVector3( f,  f,  f);
  vert[3].m_vecPos:= D3DXVector3( f,  f, -f);

  m_pDevice.SetTexture(0, m_pSurf[3]);
  m_pDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, vert, SizeOf(TEnvVertex));

  // ZNeg
  vert[0].m_vecPos:= D3DXVector3(-f,  f, -f);
  vert[1].m_vecPos:= D3DXVector3(-f, -f, -f);
  vert[2].m_vecPos:= D3DXVector3( f,  f, -f);
  vert[3].m_vecPos:= D3DXVector3( f, -f, -f);

  m_pDevice.SetTexture(0, m_pSurf[4]);
  m_pDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, vert, SizeOf(TEnvVertex));

  // ZPos
  vert[0].m_vecPos:= D3DXVector3( f,  f,  f);
  vert[1].m_vecPos:= D3DXVector3( f, -f,  f);
  vert[2].m_vecPos:= D3DXVector3(-f,  f,  f);
  vert[3].m_vecPos:= D3DXVector3(-f, -f,  f);

  m_pDevice.SetTexture(0, m_pSurf[5]);
  m_pDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, vert, SizeOf(TEnvVertex));

  Result:= S_OK;
end;

const
  WATER_SHIFT           = 6;
  WATER_SIZE            = (1 shl WATER_SHIFT);
  WATER_AREA            = (WATER_SIZE * WATER_SIZE);
  WATER_MASK            = (WATER_SIZE - 1);

  WATER_SPHERE_HEIGHT   = 20.0;
  WATER_SPHERE_RADIUS2  = (35.0 * 35.0);

// #define WATER_INDEX(x, y) \
//     ((x) | ((y) << WATER_SHIFT))
function WATER_INDEX(x, y: DWord): DWord;
begin
  Result:= x or (y shl WATER_SHIFT);
end;

// #define WATER_INDEX_WRAP(x, y) \
//     (((x) & WATER_MASK) | (((y) & WATER_MASK) << WATER_SHIFT))
function WATER_INDEX_WRAP(x, y: DWord): DWord;
begin
  Result:= (x and WATER_MASK) or ((y and WATER_MASK) shl WATER_SHIFT);
end;

// #if defined(_X86) && !defined(_WIN64)
function f2i(flt: Single): Integer;
{asm
  fld flt
  fistp eax }
begin
 Result:= Trunc(flt);
end;


//////////////////////////////////////////////////////////////////////////////
// Types /////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

type
  PWaterVertex = ^TWaterVertex;
  TWaterVertex= packed record
    m_vecPos:     TD3DXVector3;
    m_vecNormal:  TD3DXVector3;
    m_dwDiffuse:  TD3DColor;
    m_vecTex:     TD3DXVector2;
  end;

  PCausticsVertex = ^TCausticsVertex;
  TCausticsVertex = packed record
    m_vecPos:     TD3DXVector3;
    m_dwDiffuse:  TD3DColor;
  end;
  PCausticsArray = ^TCausticsArray;
  TCausticsArray = array[0..0] of TCausticsVertex;

const
  // (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0))
  TWaterVertex_FVF = D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_DIFFUSE or D3DFVF_TEX1 or D3DFVF_TEXTUREFORMAT2;

  TCausticsVertex_FVF = D3DFVF_XYZ or D3DFVF_DIFFUSE;

//////////////////////////////////////////////////////////////////////////////
// CWater implementation /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

constructor CWater.Create;
begin
  inherited;

  m_fDepth    := 0.0;
  m_fScaleTex := 1.0;

  m_uIndices  := 0;
  m_uVertices := 0;

  m_pRefract     := nil;
  m_pSurface     := nil;

  m_pDevice      := nil;
  m_pibIndices   := nil;
  m_pvbVertices  := nil;
  m_pvbCaustics  := nil;
end;


destructor CWater.Destroy;
begin
  if Assigned(m_pSurface) then FreeMem(m_pSurface);
end;

var
  // Create refraction table
  //  static WATER_REFRACT Refract[512]; ???
  Refract: array [0..511] of TWaterRefract;

function CWater.Initialize(fSize, fDepth: Single): HResult;
var
  u: Integer;

  fCos0, f0, fSin0: Single;
  fSin1, f1, fCos1: Single;
begin
  m_fSize := fSize;
  m_fDepth := fDepth;
  m_fScaleTex := 1.0 / fSize;

  // Calculate number of vertices and indices
  m_uVertices := WATER_AREA;
  m_uIndices  := m_uVertices * 2;

  // Create refraction table
  if not Assigned(m_pRefract) then
  begin
    m_pRefract := @Refract[256];

    for u:= 0 to 255 do
    begin
      fCos0 := u / 256.0;
      f0 := ArcCos(fCos0);
      fSin0 := sin(f0);

      fSin1 := fSin0 / 1.333; // water
      f1 := ArcSin(fSin1);
      fCos1 := cos(f1);

      m_pRefract[u].fRefract := fSin0 / fSin1 * fCos1 - fCos0;
      m_pRefract[u].fRefractNorm := - fSin1 / fSin0;
      m_pRefract[u].dwDiffuse := (((($ff - u)*($ff - u)*($ff - u)) shl 8) and $ff000000);

      Refract[u] := Refract[256];
    end;
  end;

  // Create maps
  if not Assigned(m_pSurface) then
  begin
    try
      GetMem(m_pSurface, SizeOf(TWaterSurface)*WATER_AREA);
    except
      Result:= E_OUTOFMEMORY;
      Exit;
    end;
    FillChar(m_pSurface^, WATER_AREA * SizeOf(TWaterSurface), $00);
  end;

  Result:= S_OK;
end;

function CWater.OnCreateDevice(const pDevice: IDirect3DDevice9): HResult;
begin
  m_pDevice:= pDevice;
  Result:= S_OK;
end;

function CWater.OnResetDevice: HResult;
var
  pwIndices: PWord;
  uX, uZ: Longword;
  pwIndex: PWord;
  uSize: LongWord;
  u: Longword;
begin
  // Create indices
  if not Assigned(m_pibIndices) then
  begin
    Result:= m_pDevice.CreateIndexBuffer(m_uIndices * SizeOf(WORD),
               D3DUSAGE_DYNAMIC or D3DUSAGE_WRITEONLY, D3DFMT_INDEX16,
               D3DPOOL_DEFAULT, m_pibIndices, nil);
    if FAILED(Result) then Exit;

    Result:= m_pibIndices.Lock(0, m_uIndices * SizeOf(WORD), Pointer(pwIndices), D3DLOCK_DISCARD);
    if FAILED(Result) then Exit;

    // Fill in indicies
    uX:= 0; uZ:= 0;
    pwIndex:= pwIndices;

    // for(UINT uSize = WATER_SIZE; uSize != 0; uSize -= 2)
    uSize:= WATER_SIZE;
    repeat
      // Top
      // for(u = 0; u < uSize; u++)
      for u:= 0 to uSize - 1 do
      begin
        pwIndex^:= uX + uZ * WATER_SIZE;                Inc(pwIndex);
        pwIndex^:= uX + uZ * WATER_SIZE + WATER_SIZE;   Inc(pwIndex);
        Inc(uX);
      end;

      Dec(uX);
      Inc(uZ);

      // Right
      // for(u = 1; u < uSize; u++)
      for u:= 1 to uSize - 1 do
      begin
        pwIndex^:= uX + uZ * WATER_SIZE;                Inc(pwIndex);
        pwIndex^:= uX + uZ * WATER_SIZE - 1;            Inc(pwIndex);
        Inc(uZ);
      end;

      Dec(uZ);
      Dec(uX);

      // Bottom
      // for(u = 1; u < uSize; u++)
      for u:= 1 to uSize - 1 do
      begin
        pwIndex^:= uX + uZ * WATER_SIZE;                Inc(pwIndex);
        pwIndex^:= uX + uZ * WATER_SIZE - WATER_SIZE;   Inc(pwIndex);
        Dec(uX);
      end;

      Inc(uX);
      Dec(uZ);

      // Left
      // for(u = 2; u < uSize; u++)
      for u:= 2 to uSize - 1 do
      begin
        pwIndex^:= uX + uZ * WATER_SIZE;                Inc(pwIndex);
        pwIndex^:= uX + uZ * WATER_SIZE + 1;            Inc(pwIndex);
        Dec(uZ);
      end;

      Inc(uZ);
      Inc(uX);

    // for(UINT uSize = WATER_SIZE; uSize != 0; uSize -= 2) // end of FOR
    Dec(uSize, 2);
    until (uSize = 0);

    // for(pwIndex = pwIndices; pwIndex < pwIndices + m_uIndices; pwIndex++)
    pwIndex:= pwIndices;
    while LongWord(pwIndex) < LongWord(pwIndices) + m_uIndices do
    begin
      if (pwIndex^ >= m_uVertices) then pwIndex^:= 0;
      Inc(pwIndex);
    end;

    m_pibIndices.Unlock;
  end;


  // Create vertices
  if not Assigned(m_pvbVertices) then
  begin
    Result:= m_pDevice.CreateVertexBuffer(m_uVertices * SizeOf(TWaterVertex),
        D3DUSAGE_WRITEONLY or D3DUSAGE_DYNAMIC, TWaterVertex_FVF, D3DPOOL_DEFAULT,
        m_pvbVertices, nil);
    if FAILED(Result) then Exit;
  end;


  // Create caustics
  if not Assigned(m_pvbCaustics) then
  begin
    Result:= m_pDevice.CreateVertexBuffer(m_uVertices * SizeOf(TCausticsVertex),
        D3DUSAGE_WRITEONLY or D3DUSAGE_DYNAMIC, TCausticsVertex_FVF, D3DPOOL_DEFAULT,
        m_pvbCaustics, nil);
    if FAILED(Result) then Exit;
  end;

  Result:= S_OK;
end;

function CWater.OnLostDevice: HResult;
begin
  m_pibIndices:= nil;
  m_pvbVertices:= nil;
  m_pvbCaustics:= nil;

  Result:= S_OK;
end;

function CWater.OnDestroyDevice: HResult;
begin
  m_pDevice:= nil;
  Result:= S_OK;
end;

function CWater.Drop: HResult;
var
  uX, uY: LongWord;
const
  RAND_MAX = $7FFF;
begin
  uX:= Random(RAND_MAX);
  uY:= Random(RAND_MAX);

  with m_pSurface[WATER_INDEX_WRAP(uX, uY)] do fVelocity:= fVelocity - 0.25;

  with m_pSurface[WATER_INDEX_WRAP(uX - 1, uY)] do fVelocity:= fVelocity - 0.125;
  with m_pSurface[WATER_INDEX_WRAP(uX + 1, uY)] do fVelocity:= fVelocity - 0.125;
  with m_pSurface[WATER_INDEX_WRAP(uX, uY - 1)] do fVelocity:= fVelocity - 0.125;
  with m_pSurface[WATER_INDEX_WRAP(uX, uY + 1)] do fVelocity:= fVelocity - 0.125;

  with m_pSurface[WATER_INDEX_WRAP(uX - 1, uY - 1)] do fVelocity:= fVelocity - 0.0625;
  with m_pSurface[WATER_INDEX_WRAP(uX + 1, uY + 1)] do fVelocity:= fVelocity - 0.0625;
  with m_pSurface[WATER_INDEX_WRAP(uX + 1, uY - 1)] do fVelocity:= fVelocity - 0.0625;
  with m_pSurface[WATER_INDEX_WRAP(uX - 1, uY + 1)] do fVelocity:= fVelocity - 0.0625;

  Result:= S_OK;
end;

const
  TABLE_SIZE = 8;
var
  Table: array[0..TABLE_SIZE - 1] of DWord = (0, 0, 0, 0, 0, 0, 0, 0);

function CWater.Update(vecPos, vecLight: TD3DXVector3; bCalcCaustics: BOOL): HResult;
var
  uXN, uX, uXP, uY, uYN, uY0, uYP: Integer; 
  pSurface: ^TWaterSurface;
  fX, fZ, fDesire: Single;

  pVertices, pVertex, pVertexLim: PWaterVertex;
  vec, vecP, vecN: TD3DXVector3;

  fInc, f: Single;

  pRefract: ^TWaterRefract;
  vecD: TD3DXVector3;
  v1, v2: TD3DXVector3; // temporary
  fC: Single;

  fB, fD, fScale: Single;

  pCaustics, pCaustic: PCausticsVertex;
  u: LongWord;

  fArea: Single;

 function CausOfs(pCaustic: PCausticsVertex; Offset: Integer): PCausticsVertex;
 begin
   Result:= PCausticsVertex(Integer(pCaustic) + Offset*SizeOf(TCausticsVertex));
 end;
begin
  // Compute desired height
  m_fAvgHeight:= 0.0;
  pSurface:= @m_pSurface[0];

  uYN:= WATER_AREA - WATER_SIZE;
  uY0:= 0;
  uYP:= WATER_SIZE;

  repeat
    uXN := WATER_SIZE - 1;
    uX  := 0;
    uXP := 1;

    repeat
      // Bowl
      fX:= 1.0*(uX - (WATER_SIZE shr 1));
      fZ:= 1.0*((uY0 shr WATER_SHIFT) - (WATER_SIZE shr 1));

      if (fX * fX + fZ * fZ) < (WATER_SPHERE_RADIUS2 - WATER_SPHERE_HEIGHT * WATER_SPHERE_HEIGHT) then
      begin
        fDesire:=
          (m_pSurface[uXN + uYN].fHeight +
           m_pSurface[uXP + uYN].fHeight +
           m_pSurface[uXN + uYP].fHeight +
           m_pSurface[uXP + uYP].fHeight) * (1.0 / 12.0)
           +
          (m_pSurface[uX  + uYN].fHeight +
           m_pSurface[uXN + uY0].fHeight +
           m_pSurface[uXP + uY0].fHeight +
           m_pSurface[uX  + uYP].fHeight) * (2.0 / 12.0);
      end else
      begin
        fDesire:= 0.0;
        pSurface.fHeight:= 0.0;
        pSurface.fVelocity:= 0.0;
      end;


      // Update velocity
      if (pSurface.fVelocity > 0.01) or (pSurface.fVelocity  < -0.01) then
        pSurface.fVelocity:= pSurface.fVelocity * 0.99;

      pSurface.fVelocity:= pSurface.fVelocity + 0.25 * (fDesire - pSurface.fHeight);
      m_fAvgHeight:= m_fAvgHeight + pSurface.fHeight + pSurface.fVelocity;

      Inc(pSurface);

      uXN := uX;
      uX  := uXP;
      uXP := (uXP + 1) and WATER_MASK;
    // while (uX);
    until (uX = 0);

    uYN:= uY0;
    uY0:= uYP;
    uYP:= (uYP + WATER_SIZE) and (WATER_MASK shl WATER_SHIFT);
  // while(uY0);
  until (uY0 = 0);

  m_fAvgHeight:= m_fAvgHeight / m_uVertices;


  // Calculate surface normals
  Result:= m_pvbVertices.Lock(0, m_uVertices * SizeOf(TWaterVertex), Pointer(pVertices), 0);
  if FAILED(Result) then Exit;

  pVertex:= pVertices;
  pVertexLim:= Pointer(DWord(pVertex) + m_uVertices);
  pSurface:= Pointer(m_pSurface);

  fInc:= m_fSize / (WATER_SIZE - 1);
  fZ:= m_fSize * -0.5;

  uYN := WATER_AREA - WATER_SIZE;
  uY0 := 0;
  uYP := WATER_SIZE;

  repeat
    fX  := m_fSize * -0.5;

    uXN := WATER_SIZE - 1;
    uX  := 0;
    uXP := 1;

    repeat
      // Update position and normal
      vecP.x:= fX;
      pSurface.fHeight:= pSurface.fHeight + pSurface.fVelocity - m_fAvgHeight;
      vecP.y:= pSurface.fHeight;
      vecP.z:= fZ;

      f:= m_pSurface[uXN + uYN].fHeight - m_pSurface[uXP + uYP].fHeight;        vecN.x:= f; vecN.z:= f;
      f:= m_pSurface[uX  + uYN].fHeight - m_pSurface[uX  + uYP].fHeight;        vecN.z:= vecN.z + f;
      f:= m_pSurface[uXP + uYN].fHeight - m_pSurface[uXN + uYP].fHeight;        vecN.x:= vecN.x - f; vecN.z:= vecN.z + f;
      f:= m_pSurface[uXN + uY0].fHeight - m_pSurface[uXP + uY0].fHeight;        vecN.x:= vecN.x + f;

      vecN.y:= 1.0;
      D3DXVec3Normalize(vecN, vecN);

      Inc(pSurface);

      // Update texture coords and diffuse based upon refraction
      D3DXVec3Subtract(vec, vecPos, vecP);
      D3DXVec3Normalize(vec, vec);

      // pRefract = m_pRefract + f2i(D3DXVec3Dot(&vec, &vecN) * 255.0f);
      pRefract:= @m_pRefract[0];
      Inc(pRefract, f2i(D3DXVec3Dot(vec, vecN) * 255));

      pVertex.m_vecPos:= vecP;
      pVertex.m_vecNormal:= vecN;
      pVertex.m_dwDiffuse:= pRefract.dwDiffuse;

      // Bowl
      // vecD:= (vecN * pRefract.fRefract + vec) * pRefract.fRefractNorm;
      D3DXVec3Scale(vecD,
        D3DXVec3Add(v1,
          D3DXVec3Scale(v2, vecN, pRefract.fRefract)^,
          vec)^,
        pRefract.fRefractNorm);
      vecP.y:= vecP.y - WATER_SPHERE_HEIGHT;

      fC:= D3DXVec3Dot(vecP, vecP) - WATER_SPHERE_RADIUS2;

      if (fC < 0.0) then
      begin
        fB:= D3DXVec3Dot(vecD, vecP) * 2.0;
        fD:= fB * fB - 4.0 * fC;
        fScale:= (-fB + sqrt(fD)) * 0.5;

        pVertex.m_vecTex.x:= (vecD.x * fScale + vecP.x) * m_fScaleTex + 0.5;
        pVertex.m_vecTex.y:= (vecD.z * fScale + vecP.z) * m_fScaleTex + 0.5;
      end else
      begin
        pVertex.m_vecTex.x:= vecP.x * m_fScaleTex + 0.5;
        pVertex.m_vecTex.y:= vecP.z * m_fScaleTex + 0.5;
      end;

      Inc(pVertex);
      fX:= fX + fInc;

      uXN:= uX;
      uX := uXP;
      uXP:= (uXP + 1) and WATER_MASK;
    // while(uX);
    until (uX = 0);

    fZ:= fZ + fInc;

    uYN:= uY0;
    uY0:= uYP;
    uYP:= (uYP + WATER_SIZE) and (WATER_MASK shl WATER_SHIFT);
  // while(uY0);
  until (uY0 = 0);


  // Calculate caustics
  if (bCalcCaustics) then
  begin
    Result:= m_pvbCaustics.Lock(0, m_uVertices * SizeOf(TCausticsVertex), Pointer(pCaustics), 0);
    if FAILED(Result) then Exit;


    if (Table[0] = 0) then
    begin
      for u:= 0 to TABLE_SIZE - 1 do
        Table[u]:= ($40 div (u + 1)) * $00010101;
    end;

    pVertex  := pVertices;
    pCaustic := pCaustics;

    for uY:= 0 to WATER_SIZE - 1 do
    begin
      for uX:= 0 to WATER_SIZE - 1 do
      begin
        // pRefract = m_pRefract + f2i(pVertex->m_vecNormal.y * 255.0f);
        pRefract:= @m_pRefract[0];
        Inc(pRefract, f2i(pVertex.m_vecNormal.y * 255.0));

        // Bowl
        // vecD = (pVertex->m_vecNormal * pRefract->fRefract + vecLight) * pRefract->fRefractNorm;
        D3DXVec3Scale(vecD,
          D3DXVec3Add(v1,
            D3DXVec3Scale(v2, pVertex.m_vecNormal, pRefract.fRefract)^,
            vecLight)^,
          pRefract.fRefractNorm);

        vecP:= pVertex.m_vecPos;
        vecP.y:= vecP.y - WATER_SPHERE_HEIGHT;

        fC:= D3DXVec3Dot(vecP, vecP) - WATER_SPHERE_RADIUS2;

        if (fC < 0.0) then
        begin
          fB:= D3DXVec3Dot(vecD, vecP) * 2.0;
          fD:= fB * fB - 4.0 * fC;
          fScale:= (-fB + sqrt(fD)) * 0.5;

          pCaustic.m_vecPos.x:= vecD.x * fScale + vecP.x;
          pCaustic.m_vecPos.y:= 0.0;
          pCaustic.m_vecPos.z:= vecD.z * fScale + vecP.z;
        end else
        begin
          // Idle water outside of sphere 
          pCaustic.m_vecPos.x:= vecP.x;
          pCaustic.m_vecPos.y:= 0.0;
          pCaustic.m_vecPos.z:= vecP.z;
        end;

        if (uX<>0) and (uY<>0) then
        begin
          fArea:= (CausOfs(pCaustic, -WATER_SIZE - 1).m_vecPos.x - pCaustic.m_vecPos.x) *
                  (CausOfs(pCaustic, -WATER_SIZE    ).m_vecPos.z - pCaustic.m_vecPos.z) -
                  (CausOfs(pCaustic, -WATER_SIZE - 1).m_vecPos.z - pCaustic.m_vecPos.z) *
                  (CausOfs(pCaustic, -WATER_SIZE    ).m_vecPos.x - pCaustic.m_vecPos.x);

          if(fArea < 0.0) then
              fArea:= -fArea;

          u:= f2i(fArea * fArea * 4.0);
          if u < TABLE_SIZE
            then pCaustic.m_dwDiffuse:= Table[u]
            else pCaustic.m_dwDiffuse:= 0;
        end;

        Inc(pCaustic);
        Inc(pVertex);
      end;

      // pCaustic[-WATER_SIZE].m_dwDiffuse = pCaustic[-1].m_dwDiffuse;
      CausOfs(pCaustic, -WATER_SIZE).m_dwDiffuse:=
        CausOfs(pCaustic, -1).m_dwDiffuse;
    end;

    for uX:= 0 to WATER_SIZE - 1 do
      CausOfs(pCaustics, uX).m_dwDiffuse:= CausOfs(pCaustics, uX + (WATER_AREA - WATER_SIZE)).m_dwDiffuse;

    m_pvbCaustics.Unlock;
  end;

  m_pvbVertices.Unlock;

  Result:= S_OK;
end;

function CWater.DrawCaustics: HResult;
begin
  Result:= m_pDevice.SetFVF(TCausticsVertex_FVF);
  if FAILED(Result) then Exit;

  Result:= m_pDevice.SetStreamSource(0, m_pvbCaustics, 0, SizeOf(TCausticsVertex));
  if FAILED(Result) then Exit;

  Result:= m_pDevice.SetIndices(m_pibIndices);
  if FAILED(Result) then Exit;

  Result:= m_pDevice.DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0, 0, m_uVertices, 0, m_uIndices -2);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

function CWater.DrawSurface: HResult;
begin
  Result:= m_pDevice.SetFVF(TWaterVertex_FVF);
  if FAILED(Result) then Exit;

  Result:= m_pDevice.SetStreamSource(0, m_pvbVertices, 0, SizeOf(TWaterVertex));
  if FAILED(Result) then Exit;

  Result:= m_pDevice.SetIndices(m_pibIndices);
  if FAILED(Result) then Exit;

  Result:= m_pDevice.DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0, 0, m_uVertices, 0, m_uIndices -2);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;


end.

